@@ -20,18 +20,21 @@
 /* private function prototypes */
 static int validxmlent(const char *xml);
 static char *uniqueid(const char *prefix);
-static int validid(const char *id);
+static int validid(const char *id, const char *prefix);
 static char *datestr(time_t t);
 static int validdate(const char *date);
 static const char *wdayname(int wday);
 static const char *monname(int mon);
 static int strisprintable(const char *str);
+static const char *skiplabel(const char *str);
+static void wikicat(CBDATUM *buf, const char *text, int trim);
 static char *headtohtml(ARTICLE *art);
 static char *bodytohtml(ARTICLE *art, const char *systemuri, int num);
 static char *tailtohtml(ARTICLE *art, int num);
 const char *artsubclass(const char *flags);
 const char *ressubclass(const char *flags);
 static const char *hashclass(const char *prefix, const char *str);
+static const char *latexverbstr(void);
 
 
 
@@ -122,6 +125,10 @@ int rbbsremove(const char *name){
   vlremove(path);
   sprintf(path, "%s/%s", name, MDIDXNAME);
   vlremove(path);
+  sprintf(path, "%s/%s", name, FILEDBNAME);
+  crremove(path);
+  sprintf(path, "%s/%s", name, FILEIDXNAME);
+  dpremove(path);
   if(rmdir(name) == -1) return FALSE;
   return TRUE;
 }
@@ -174,6 +181,16 @@ int rbbsoptimize(RBBS *rbbs){
       rbbs->errmsg = cbsprintf("DB-Error: the database for modification date is broken");
     return FALSE;
   }
+  if(!croptimize(rbbs->filedb, -1)){
+    if(!rbbs->errmsg)
+      rbbs->errmsg = cbsprintf("DB-Error: the database for file entities is broken");
+    return FALSE;
+  }
+  if(!dpoptimize(rbbs->fileidx, -1)){
+    if(!rbbs->errmsg)
+      rbbs->errmsg = cbsprintf("DB-Error: the database for file attributes is broken");
+    return FALSE;
+  }
   return TRUE;
 }
 
@@ -265,6 +282,11 @@ int postarticle(RBBS *rbbs, ARTICLE *art, const char *id){
   int err;
   assert(rbbs && art);
   if(id && !deletearticle(rbbs, id)) return FALSE;
+  if(!id && dpvsiz(rbbs->artdb, art->id, -1) != -1){
+    if(!rbbs->errmsg)
+      rbbs->errmsg = cbsprintf("DB-Error: the ID of the article is duplicated");
+    return FALSE;
+  }
   oid = id ? id : art->id;
   xml = arttoxml(art);
   err = FALSE;
@@ -335,6 +357,7 @@ int deletearticle(RBBS *rbbs, const char *id){
 char *getattribute(RBBS *rbbs, const char *name){
   assert(rbbs && name);
   if(cbstrfwmatch(name, AIDPREFIX)) return NULL;
+  if(cbstrfwmatch(name, FIDPREFIX)) return NULL;
   return dpget(rbbs->artdb, name, -1, 0, -1, NULL);
 }
 
@@ -343,6 +366,7 @@ char *getattribute(RBBS *rbbs, const char *name){
 int setattribute(RBBS *rbbs, const char *name, const char *value){
   assert(rbbs && name);
   if(cbstrfwmatch(name, AIDPREFIX)) return FALSE;
+  if(cbstrfwmatch(name, FIDPREFIX)) return FALSE;
   if(value){
     return dpput(rbbs->artdb, name, -1, value, -1, DP_DOVER);
   } else {
@@ -352,13 +376,13 @@ int setattribute(RBBS *rbbs, const char *name, const char *value){
 
 
 /* Save a file into a database. */
-int savefile(RBBS *rbbs, const char *password, const char *subject, const char *author,
-             const char *cdate, const char *ptr, int size){
+int savefile(RBBS *rbbs, const char *id, const char *password, const char *subject,
+             const char *author, const char *cdate, const char *ptr, int size){
   CBMAP *attrs;
-  char *id, *mbuf, numbuf[NUMBUFSIZ];
+  char *myid, *mbuf, numbuf[NUMBUFSIZ];
   int err, mbsiz;
   assert(rbbs && password && subject && author && cdate && ptr && size >= 0);
-  id = uniqueid(FIDPREFIX);
+  myid = id ? cbmemdup(id, -1) : uniqueid(FIDPREFIX);
   err = FALSE;
   attrs = cbmapopenex(1);
   cbmapput(attrs, "password", -1, makecrypt(password), -1, TRUE);
@@ -368,11 +392,20 @@ int savefile(RBBS *rbbs, const char *password, const char *subject, const char *
   sprintf(numbuf, "%d", size);
   cbmapput(attrs, "size", -1, numbuf, -1, TRUE);
   mbuf = cbmapdump(attrs, &mbsiz);
-  if(!dpput(rbbs->fileidx, id, -1, mbuf, mbsiz, FALSE)) err = TRUE;
-  if(!crput(rbbs->filedb, id, -1, ptr, size, FALSE)) err = TRUE;
+  if(!dpput(rbbs->fileidx, myid, -1, mbuf, mbsiz, DP_DOVER)){
+    if(!rbbs->errmsg)
+      rbbs->errmsg = cbsprintf("DB-Error: the database for files is broken or not writable");
+    err = TRUE;
+  }
+  if(!crput(rbbs->filedb, myid, -1, ptr, size, DP_DOVER)){
+    if(!rbbs->errmsg)
+      rbbs->errmsg = cbsprintf("DB-Error: the database for file attriubtes"
+                               " is broken or not writable");
+    err = TRUE;
+  }
   free(mbuf);
   cbmapclose(attrs);
-  free(id);
+  free(myid);
   return err ? FALSE : TRUE;
 }
 
@@ -384,7 +417,10 @@ char *loadfile(RBBS *rbbs, const char *id, int *sp, CBMAP *attrs){
   char *mbuf;
   int mbsiz;
   assert(rbbs && id);
-  if(!(mbuf = dpget(rbbs->fileidx, id, -1, 0, -1, &mbsiz))) return NULL;
+  if(!(mbuf = dpget(rbbs->fileidx, id, -1, 0, -1, &mbsiz))){
+    if(!rbbs->errmsg) rbbs->errmsg = cbsprintf("DB-Error: there is no file for \"%s\"", id);
+    return NULL;
+  }
   if(attrs){
     map = cbmapload(mbuf, mbsiz);
     if(!(tmp = cbmapget(map, "password", -1, NULL))) tmp = "";
@@ -398,8 +434,8 @@ char *loadfile(RBBS *rbbs, const char *id, int *sp, CBMAP *attrs){
     if(!(tmp = cbmapget(map, "size", -1, NULL))) tmp = "";
     cbmapput(attrs, "size", -1, tmp, -1, TRUE);
     cbmapclose(map);
-    free(mbuf);
   }
+  free(mbuf);
   if(!sp) return NULL;
   return crget(rbbs->filedb, id, -1, 0, -1, sp);
 }
@@ -410,7 +446,10 @@ int deletefile(RBBS *rbbs, const char *id){
   int err;
   assert(rbbs && id);
   err = FALSE;
-  if(!dpout(rbbs->fileidx, id, -1)) err = TRUE;
+  if(!dpout(rbbs->fileidx, id, -1)){
+    if(!rbbs->errmsg) rbbs->errmsg = cbsprintf("DB-Error: there is no file for \"%s\"", id);
+    err = TRUE;
+  }
   if(!crout(rbbs->filedb, id, -1)) err = TRUE;
   return err ? FALSE : TRUE;
 }
@@ -432,7 +471,14 @@ CBLIST *listfiles(RBBS *rbbs){
 }
 
 
-/* Make a object of article from an XML string. */
+/* Get the number of files in a database. */
+int numfiles(RBBS *rbbs){
+  assert(rbbs);
+  return crrnum(rbbs->filedb);
+}
+
+
+/* Make an object of an article from an XML string. */
 ARTICLE *makearticle(const char *xml){
   ARTICLE *art;
   CBLIST *elems, *stack;
@@ -481,8 +527,9 @@ ARTICLE *makearticle(const char *xml){
         if(!pname || strcmp(name, pname)){
           art->errmsg = cbsprintf("Invalid-XML: an end tag \"%s\" is mismatching", name);
         } else {
-          if(bstart >= 0 && (!strcmp(pname, "topic") || !strcmp(pname, "para") ||
-                             !strcmp(pname, "asis") || !strcmp(pname, "list"))){
+          if(bstart >= 0 &&
+             (!strcmp(pname, "topic") || !strcmp(pname, "subtopic") || !strcmp(pname, "para") ||
+              !strcmp(pname, "asis") || !strcmp(pname, "list"))){
             for(j = bstart; j <= i; j++){
               celem = cblistval(elems, j, NULL);
               cbdatumcat(cont, celem, -1);
@@ -630,6 +677,14 @@ ARTICLE *makearticle(const char *xml){
           } else {
             bstart = i;
           }
+        } else if(!strcmp(name, "subtopic")){
+          if(!pname || strcmp(pname, "body")){
+            art->errmsg = cbsprintf("Invalid-XML: \"subtopic\" must be a child of \"body\"");
+          } else if(cbstrbwmatch(elem, "/>")){
+            art->errmsg = cbsprintf("Invalid-XML: \"subtopic\" must not be empty");
+          } else {
+            bstart = i;
+          }
         } else if(!strcmp(name, "para")){
           if(!pname || strcmp(pname, "body")){
             art->errmsg = cbsprintf("Invalid-XML: \"para\" must be a child of \"body\"");
@@ -745,10 +800,10 @@ ARTICLE *makearticle(const char *xml){
         if(!pname) pname = "";
         if(strcmp(pname, "flags") && strcmp(pname, "language") && strcmp(pname, "category") &&
            strcmp(pname, "subject") && strcmp(pname, "author") && strcmp(pname, "cdate") &&
-           strcmp(pname, "mdate") && strcmp(pname, "topic") && strcmp(pname, "para") &&
-           strcmp(pname, "asis") && strcmp(pname, "item") && strcmp(pname, "link") &&
-           strcmp(pname, "emp") && strcmp(pname, "cite") && strcmp(pname, "ins") &&
-           strcmp(pname, "del") && strcmp(pname, "response")){
+           strcmp(pname, "mdate") && strcmp(pname, "topic") && strcmp(pname, "subtopic") &&
+           strcmp(pname, "para") && strcmp(pname, "asis") && strcmp(pname, "item") &&
+           strcmp(pname, "link") && strcmp(pname, "emp") && strcmp(pname, "cite") &&
+           strcmp(pname, "ins") && strcmp(pname, "del") && strcmp(pname, "response")){
           art->errmsg =
             cbsprintf("Invalid-XML: a printable character must not be included in a block");
         }
@@ -758,7 +813,7 @@ ARTICLE *makearticle(const char *xml){
   }
   if(!art->id){
     art->id = uniqueid(AIDPREFIX);
-  } else if(!validid(art->id)){
+  } else if(!validid(art->id, AIDPREFIX)){
     if(!art->errmsg) art->errmsg = cbsprintf("Invalid-XML: the ID is invalid");
   }
   if(!art->password) art->password = cbmemdup("", 0);
@@ -798,6 +853,143 @@ ARTICLE *makearticle(const char *xml){
 }
 
 
+/* Make an object of an article from a wiki-style string. */
+ARTICLE *makeartfromwiki(const char *wiki){
+  const char *delimsrc = ">>>>>>>>>>>>>>>>";
+  const char *topicstr = "!";
+  const char *subtpcstr = "!!";
+  const char *itemstr = "*";
+  const char *asisstr = "|";
+  const char *graphstr = "@graph:";
+  const char *breakstr = "----";
+  ARTICLE *art;
+  CBDATUM *buf;
+  CBLIST *lines;
+  const char *id, *password, *flags, *language, *category, *subject, *author, *cdate, *mdate;
+  const char *delim, *cline, *pline, *nline;
+  char *rtext;
+  int i, bs, be;
+  assert(wiki);
+  rtext = strnormalize(wiki, FALSE, FALSE);
+  buf = cbdatumopen("", 0);
+  lines = cbsplit(rtext, -1, "\n");
+  id = "";
+  password = "";
+  flags = "";
+  language = "";
+  category = "";
+  subject = "";
+  author = "";
+  cdate = "";
+  mdate = "";
+  delim = delimsrc;
+  bs = 0;
+  for(i = 0; i < cblistnum(lines); i++){
+    cline = cblistval(lines, i, NULL);
+    if(cbstrfwmatch(cline, "id:")){
+      id = skiplabel(cline);
+    } else if(cbstrfwmatch(cline, "password:")){
+      password = skiplabel(cline);
+    } else if(cbstrfwmatch(cline, "flags:")){
+      flags = skiplabel(cline);
+    } else if(cbstrfwmatch(cline, "language:")){
+      language = skiplabel(cline);
+    } else if(cbstrfwmatch(cline, "category:")){
+      category = skiplabel(cline);
+    } else if(cbstrfwmatch(cline, "subject:")){
+      subject = skiplabel(cline);
+    } else if(cbstrfwmatch(cline, "author:")){
+      author = skiplabel(cline);
+    } else if(cbstrfwmatch(cline, "cdate:")){
+      cdate = skiplabel(cline);
+    } else if(cbstrfwmatch(cline, "mdate:")){
+      mdate = skiplabel(cline);
+    } else if(cbstrfwmatch(cline, delim)){
+      bs = i + 1;
+      delim = cline;
+      break;
+    }
+  }
+  xsprintf(buf, "<article");
+  if(id[0] != '\0') xsprintf(buf, " id=\"%@\"", id);
+  if(password[0] != '\0') xsprintf(buf, " password=\"%@\"", password);
+  xsprintf(buf, ">\n");
+  xsprintf(buf, "<head>\n");
+  xsprintf(buf, "<flags>%@</flags>\n", flags);
+  xsprintf(buf, "<language>%@</language>\n", language);
+  xsprintf(buf, "<category>%@</category>\n", category);
+  xsprintf(buf, "<subject>%@</subject>\n", subject);
+  xsprintf(buf, "<author>%@</author>\n", author);
+  xsprintf(buf, "<cdate>%@</cdate>\n", cdate);
+  xsprintf(buf, "<mdate>%@</mdate>\n", mdate);
+  xsprintf(buf, "</head>\n");
+  xsprintf(buf, "<body>\n");
+  be = cblistnum(lines);
+  for(i = bs; i < cblistnum(lines); i++){
+    if(!strcmp(cblistval(lines, i, NULL), delim)){
+      be = i;
+      break;
+    }
+  }
+  for(i = bs; i < be; i++){
+    pline = i > bs ? cblistval(lines, i - 1, NULL) : "";
+    cline = cblistval(lines, i, NULL);
+    nline = i < be - 1 ? cblistval(lines, i + 1, NULL) : "";
+    if(cline[0] != '\0'){
+      if(cbstrfwmatch(cline, subtpcstr)){
+        xsprintf(buf, "<subtopic>");
+        wikicat(buf, cline + strlen(subtpcstr), TRUE);
+        xsprintf(buf, "</subtopic>\n");
+      } else if(cbstrfwmatch(cline, topicstr)){
+        xsprintf(buf, "<topic>");
+        wikicat(buf, cline + strlen(topicstr), TRUE);
+        xsprintf(buf, "</topic>\n");
+      } else if(cbstrfwmatch(cline, itemstr)){
+        if(!cbstrfwmatch(pline, itemstr)) xsprintf(buf, "<list>\n");
+        xsprintf(buf, "<item>");
+        wikicat(buf, cline + strlen(itemstr), TRUE);
+        xsprintf(buf, "</item>\n");
+        if(!cbstrfwmatch(nline, itemstr)) xsprintf(buf, "</list>\n");
+      } else if(cbstrfwmatch(cline, asisstr)){
+        if(!cbstrfwmatch(pline, asisstr)) xsprintf(buf, "<asis>");
+        xsprintf(buf, "%s\n", cline + strlen(asisstr));
+        if(!cbstrfwmatch(nline, asisstr)) xsprintf(buf, "</asis>\n");
+      } else if(cbstrfwmatch(cline, graphstr)){
+        cline += strlen(graphstr);
+        while(*cline == ' '){
+          cline++;
+        }
+        xsprintf(buf, "<graph data=\"%@\"/>\n", cline);
+      } else if(!strcmp(cline, breakstr)){
+        xsprintf(buf, "<break/>\n");
+      } else {
+        xsprintf(buf, "<para>");
+        wikicat(buf, cline, TRUE);
+        xsprintf(buf, "</para>\n");
+      }
+    }
+  }
+  xsprintf(buf, "</body>\n");
+  xsprintf(buf, "<tail>\n");
+  for(i = be + 1; i < cblistnum(lines) - 3; i += 4){
+    flags = cblistval(lines, i, NULL);
+    author = cblistval(lines, i + 1, NULL);
+    cdate = cblistval(lines, i + 2, NULL);
+    cline = cblistval(lines, i + 3, NULL);
+    if(author[0] != '\0' && cdate[0] != '\0' && validdate(cdate) && cline != '\0')
+      xsprintf(buf, "<response flags=\"%@\" author=\"%@\" cdate=\"%@\">%@</response>\n",
+               flags, author, cdate, cline);
+  }
+  xsprintf(buf, "</tail>\n");
+  xsprintf(buf, "</article>\n");
+  art = makearticle(cbdatumptr(buf));
+  cblistclose(lines);
+  cbdatumclose(buf);
+  free(rtext);
+  return art;
+}
+
+
 /* Release resources of an article object. */
 void freearticle(ARTICLE *art){
   assert(art);
@@ -973,8 +1165,9 @@ char *arttoxml(ARTICLE *art){
         name = cbmapget(attrs, "", -1, NULL);
         if(elem[1] == '/'){
           xsprintf(buf, "</%@>", name);
-          if(!strcmp(name, "topic") || !strcmp(name, "para") || !strcmp(name, "asis") ||
-             !strcmp(name, "list") || !strcmp(name, "item")) xsprintf(buf, "\n");
+          if(!strcmp(name, "topic") || !strcmp(name, "subtopic") || !strcmp(name, "para") ||
+             !strcmp(name, "asis") || !strcmp(name, "list") || !strcmp(name, "item"))
+            xsprintf(buf, "\n");
           if(!strcmp(name, "list")){
             inlist = FALSE;
           } else if(!strcmp(name, "item")){
@@ -1186,6 +1379,414 @@ char *arttoatom(ARTICLE *art, const char *systemuri, int bnum, int rnum){
 }
 
 
+/* Serialize an article into a string for LaTeX. */
+char *arttolatex(ARTICLE *art, int header){
+  CBDATUM *buf;
+  CBLIST *elems;
+  CBMAP *attrs;
+  const char *vstr, *blk, *elem, *name, *attr, *tmp;
+  char *rtext, *ntext, *ptext, *url, *flags, *author, *cdate, *text, *etext;
+  int i, j, asis, del, inlist, initem, tlf, len;
+  assert(art);
+  buf = cbdatumopen("", 0);
+  if(header){
+    rtext = strtolatex(artsubject(art));
+    xsprintf(buf, "\\chapter{%s}\n\n", rtext);
+    free(rtext);
+    xsprintf(buf, "\\begin{flushright}\n");
+    xsprintf(buf, "\\vspace{-3.8em}\n");
+    rtext = strtolatex(artauthor(art));
+    xsprintf(buf, "\\noindent --- {\\it %s}\\\\[0.4em]\n", rtext);
+    free(rtext);
+    rtext = strtolatex(artcdate(art));
+    xsprintf(buf, "{\\footnotesize %s}\\\\\n", rtext);
+    free(rtext);
+    rtext = strtolatex(artmdate(art));
+    xsprintf(buf, "{\\footnotesize %s}\n", rtext);
+    free(rtext);
+    xsprintf(buf, "\\end{flushright}\n");
+    xsprintf(buf, "\\vspace{0.2em}\n\n");
+  }
+  vstr = NULL;
+  for(i = 0; i < cblistnum(art->body); i++){
+    if(strstr(cblistval(art->body, i, NULL), "\\end{verbatim}")){
+      vstr = latexverbstr();
+      break;
+    }
+  }
+  if(!vstr) vstr = "verbatim";
+  asis = FALSE;
+  del = FALSE;
+  inlist = FALSE;
+  initem = FALSE;
+  tlf = FALSE;
+  url = NULL;
+  for(i = 0; i < cblistnum(art->body); i++){
+    blk = cblistval(art->body, i, NULL);
+    elems = cbxmlbreak(blk, FALSE);
+    for(j = 0; j < cblistnum(elems); j++){
+      elem = cblistval(elems, j, NULL);
+      if(elem[0] == '<'){
+        attrs = cbxmlattrs(elem);
+        name = cbmapget(attrs, "", -1, NULL);
+        if(!strcmp(name, "topic")){
+          if(elem[1] == '/'){
+            xsprintf(buf, "}\n\n");
+          } else {
+            xsprintf(buf, "\\section{");
+          }
+        } else if(!strcmp(name, "subtopic")){
+          if(elem[1] == '/'){
+            xsprintf(buf, "}\n\n");
+          } else {
+            xsprintf(buf, "\\subsection{");
+          }
+        } else if(!strcmp(name, "para")){
+          if(elem[1] == '/'){
+            xsprintf(buf, "\n\n");
+          }
+        } else if(!strcmp(name, "asis")){
+          if(elem[1] == '/'){
+            if(!tlf) xsprintf(buf, "\n");
+            xsprintf(buf, "\\end{%s}\n", vstr);
+            xsprintf(buf, "\\end{minipage}\n");
+            xsprintf(buf, "\\vspace{0.5em}\n\n");
+            asis = FALSE;
+            tlf = FALSE;
+          } else {
+            xsprintf(buf, "\\vspace{0.5em}\n");
+            xsprintf(buf, "\\noindent \\hspace{0.8em}\n");
+            xsprintf(buf, "\\begin{minipage}[t]{\\textwidth}\n");
+            xsprintf(buf, "\\begin{%s}\n", vstr);
+            asis = TRUE;
+            tlf = TRUE;
+          }
+        } else if(!strcmp(name, "list")){
+          if(elem[1] == '/'){
+            xsprintf(buf, "\\end{itemize}\n");
+            xsprintf(buf, "\\vspace{-1.0em}\n\n");
+            inlist = FALSE;
+          } else {
+            xsprintf(buf, "\\vspace{-1.0em}\n");
+            xsprintf(buf, "\\begin{itemize}\n");
+            xsprintf(buf, "\\itemsep=-0.25em\n");
+            inlist = TRUE;
+          }
+        } else if(!strcmp(name, "item")){
+          if(elem[1] == '/'){
+            xsprintf(buf, "\n");
+            initem = FALSE;
+          } else {
+            xsprintf(buf, "\\item ");
+            initem = TRUE;
+          }
+        } else if(!strcmp(name, "graph")){
+          xsprintf(buf, "\\begin{figure}[!h]\n");
+          xsprintf(buf, "\\begin{center}\n");
+          if(!(attr = cbmapget(attrs, "data", -1, NULL))) attr = "";
+          if((tmp = strrchr(attr, '/')) != NULL) attr = tmp + 1;
+          xsprintf(buf, "\\includegraphics[width=8cm]{%s}\n", attr);
+          xsprintf(buf, "\\end{center}\n");
+          xsprintf(buf, "\\end{figure}\n\n");
+        } else if(!strcmp(name, "break")){
+          xsprintf(buf, "\\vspace{1.5em}\n");
+          xsprintf(buf, "\\hrule width 0.8\\textwidth\n");
+          xsprintf(buf, "\\vspace{1.3em}\n\n");
+        } else if(!asis && !strcmp(name, "link")){
+          if(elem[1] == '/'){
+            xsprintf(buf, "}");
+            if(url){
+              ptext = strtolatex(url);
+              xsprintf(buf, "\\footnote{%s}", ptext);
+              free(ptext);
+            }
+          } else {
+            free(url);
+            if(!(attr = cbmapget(attrs, "to", -1, NULL))) attr = "";
+            url = cbmemdup(attr, -1);
+            xsprintf(buf, "\\textcolor{link}{");
+          }
+        } else if(!asis && !strcmp(name, "emp")){
+          if(elem[1] == '/'){
+            xsprintf(buf, "}");
+          } else {
+            xsprintf(buf, "{\\bf ");
+          }
+        } else if(!asis && !strcmp(name, "cite")){
+          if(elem[1] == '/'){
+            xsprintf(buf, "}");
+          } else {
+            xsprintf(buf, "{\\it ");
+          }
+        } else if(!asis && !strcmp(name, "ins")){
+          if(elem[1] == '/'){
+            xsprintf(buf, "}");
+          } else {
+            xsprintf(buf, "\\textcolor{ins}{");
+          }
+        } else if(!strcmp(name, "del")){
+          if(elem[1] == '/'){
+            if(asis){
+              del = FALSE;
+            } else {
+              xsprintf(buf, "}");
+            }
+          } else {
+            if(asis){
+              del = TRUE;
+            } else {
+              xsprintf(buf, "\\textcolor{del}{");
+            }
+          }
+        }
+        cbmapclose(attrs);
+      } else if(!del && !(inlist && !initem)){
+        rtext = cbxmlunescape(elem);
+        if(asis){
+          xsprintf(buf, "%s", rtext);
+          len = strlen(rtext);
+          tlf = len > 0 && rtext[len-1] == '\n';
+        } else {
+          ntext = strnormalize(rtext, TRUE, FALSE);
+          ptext = strtolatex(ntext);
+          xsprintf(buf, "%s", ptext);
+          free(ptext);
+          free(ntext);
+        }
+        free(rtext);
+      }
+    }
+    cblistclose(elems);
+  }
+  free(url);
+  if(cblistnum(art->tail) > 0){
+    xsprintf(buf, "\\vspace{2.2em}\n");
+    xsprintf(buf, "\\hrule width 0.95\\textwidth\n");
+    xsprintf(buf, "\\vspace{0.2em}\n\n");
+    xsprintf(buf, "\\begin{itemize}\n");
+    xsprintf(buf, "\\itemsep=-0.25em\n");
+    xsprintf(buf, "\\itemindent=-1.5em\n");
+    for(i = 0; i < cblistnum(art->tail); i++){
+      flags = NULL;
+      author = NULL;
+      cdate = NULL;
+      text = NULL;
+      blk = cblistval(art->tail, i, NULL);
+      elems = cbxmlbreak(blk, FALSE);
+      for(j = 0; j < cblistnum(elems); j++){
+        elem = cblistval(elems, j, NULL);
+        if(elem[0] == '<'){
+          attrs = cbxmlattrs(elem);
+          name = cbmapget(attrs, "", -1, NULL);
+          if(!strcmp(name, "response") && elem[1] != '/'){
+            if(!flags && (tmp = cbmapget(attrs, "flags", -1, NULL)) != NULL)
+              flags = strtolatex(tmp);
+            if(!author && (tmp = cbmapget(attrs, "author", -1, NULL)) != NULL)
+              author = strtolatex(tmp);
+            if(!cdate && (tmp = cbmapget(attrs, "cdate", -1, NULL)) != NULL)
+              cdate = strtolatex(tmp);
+          }
+          cbmapclose(attrs);
+        } else {
+          if(!text && strisprintable(elem)){
+            rtext = strnormalize(elem, TRUE, TRUE);
+            etext = cbxmlunescape(rtext);
+            text = strtolatex(etext);
+            free(etext);
+            free(rtext);
+          }
+        }
+      }
+      cblistclose(elems);
+      if(flags && author && cdate && text){
+        xsprintf(buf, "\\item[] {\\small {\\it %s} --- %s }{\\scriptsize --- %s}\n",
+                 author, text, cdate);
+      }
+      free(text);
+      free(cdate);
+      free(author);
+      free(flags);
+    }
+    xsprintf(buf, "\\end{itemize}\n\n");
+  }
+  return cbdatumtomalloc(buf, NULL);
+}
+
+
+/* Serialize an article into a Wiki-style string. */
+char *arttowiki(ARTICLE *art){
+  CBDATUM *buf;
+  CBLIST *elems;
+  CBMAP *attrs;
+  const char *blk, *elem, *name, *attr, *tmp;
+  char delim[NUMBUFSIZ*2], *rtext, *url, *flags, *author, *cdate, *text;
+  int i, j, k, asis, inlist, initem, tlf;
+  assert(art);
+  sprintf(delim, ">>>>>>>>>>>>>>>>%s%08X", art->hash,
+          (dpouterhash(art->subject, -1) * dpouterhash(art->author, -1)) & 0xffffffff);
+  buf = cbdatumopen("", 0);
+  xsprintf(buf, "id: %s\n", art->id);
+  xsprintf(buf, "password: %s\n", art->password);
+  xsprintf(buf, "flags: %s\n", art->flags);
+  xsprintf(buf, "language: %s\n", art->language);
+  xsprintf(buf, "category: %s\n", art->category);
+  xsprintf(buf, "subject: %s\n", art->subject);
+  xsprintf(buf, "author: %s\n", art->author);
+  xsprintf(buf, "cdate: %s\n", art->cdate);
+  xsprintf(buf, "mdate: %s\n", art->mdate);
+  xsprintf(buf, "%s\n\n", delim);
+  asis = FALSE;
+  inlist = FALSE;
+  initem = FALSE;
+  tlf = FALSE;
+  url = NULL;
+  for(i = 0; i < cblistnum(art->body); i++){
+    blk = cblistval(art->body, i, NULL);
+    elems = cbxmlbreak(blk, FALSE);
+    for(j = 0; j < cblistnum(elems); j++){
+      elem = cblistval(elems, j, NULL);
+      if(elem[0] == '<'){
+        attrs = cbxmlattrs(elem);
+        name = cbmapget(attrs, "", -1, NULL);
+        if(!strcmp(name, "topic")){
+          if(elem[1] == '/'){
+            xsprintf(buf, "\n\n");
+          } else {
+            xsprintf(buf, "! ");
+          }
+        } else if(!strcmp(name, "subtopic")){
+          if(elem[1] == '/'){
+            xsprintf(buf, "\n\n");
+          } else {
+            xsprintf(buf, "!! ");
+          }
+        } else if(!strcmp(name, "para")){
+          if(elem[1] == '/'){
+            xsprintf(buf, "\n\n");
+          } else {
+            xsprintf(buf, "  ");
+          }
+        } else if(!strcmp(name, "asis")){
+          if(elem[1] == '/'){
+            if(!tlf) xsprintf(buf, "\n");
+            xsprintf(buf, "\n");
+            asis = FALSE;
+            tlf = FALSE;
+          } else {
+            asis = TRUE;
+            tlf = TRUE;
+          }
+        } else if(!strcmp(name, "list")){
+          if(elem[1] == '/'){
+            xsprintf(buf, "\n");
+            inlist = FALSE;
+          } else {
+            inlist = TRUE;
+          }
+        } else if(!strcmp(name, "item")){
+          if(elem[1] == '/'){
+            xsprintf(buf, "\n");
+            initem = FALSE;
+          } else {
+            xsprintf(buf, "* ");
+            initem = TRUE;
+          }
+        } else if(!strcmp(name, "graph")){
+          if(!(attr = cbmapget(attrs, "data", -1, NULL))) attr = "";
+          xsprintf(buf, "@graph: %s\n\n", attr);
+        } else if(!strcmp(name, "break")){
+          xsprintf(buf, "----\n\n");
+        } else if(!asis && !strcmp(name, "link")){
+          if(elem[1] == '/'){
+            xsprintf(buf, "|%s]]", url ? url : "");
+          } else {
+            free(url);
+            if(!(attr = cbmapget(attrs, "to", -1, NULL))) attr = "";
+            url = cbmemdup(attr, -1);
+            xsprintf(buf, "[[");
+          }
+        } else if(!asis && !strcmp(name, "emp")){
+          if(elem[1] == '/'){
+            xsprintf(buf, "]]");
+          } else {
+            xsprintf(buf, "[[^");
+          }
+        } else if(!asis && !strcmp(name, "cite")){
+          if(elem[1] == '/'){
+            xsprintf(buf, "]]");
+          } else {
+            xsprintf(buf, "[[~");
+          }
+        } else if(!asis && !strcmp(name, "ins")){
+          if(elem[1] == '/'){
+            xsprintf(buf, "]]");
+          } else {
+            xsprintf(buf, "[[+");
+          }
+        } else if(!asis && !strcmp(name, "del")){
+          if(elem[1] == '/'){
+            xsprintf(buf, "]]");
+          } else {
+            xsprintf(buf, "[[-");
+          }
+        }
+        cbmapclose(attrs);
+      } else if(!(inlist && !initem)){
+        rtext = cbxmlunescape(elem);
+        if(tlf) cbdatumcat(buf, "|", 1);
+        for(k = 0; rtext[k] != '\0'; k++){
+          cbdatumcat(buf, rtext + k, 1);
+          if(asis && rtext[k] == '\n' && rtext[k+1] != '\0') cbdatumcat(buf, "|", 1);
+        }
+        tlf = asis && k > 0 && rtext[k-1] == '\n';
+        free(rtext);
+      }
+    }
+    cblistclose(elems);
+  }
+  free(url);
+  xsprintf(buf, "%s\n", delim);
+  for(i = 0; i < cblistnum(art->tail); i++){
+    flags = NULL;
+    author = NULL;
+    cdate = NULL;
+    text = NULL;
+    blk = cblistval(art->tail, i, NULL);
+    elems = cbxmlbreak(blk, FALSE);
+    for(j = 0; j < cblistnum(elems); j++){
+      elem = cblistval(elems, j, NULL);
+      if(elem[0] == '<'){
+        attrs = cbxmlattrs(elem);
+        name = cbmapget(attrs, "", -1, NULL);
+        if(!strcmp(name, "response") && elem[1] != '/'){
+          if(!flags && (tmp = cbmapget(attrs, "flags", -1, NULL)) != NULL)
+            flags = cbmemdup(tmp, -1);
+          if(!author && (tmp = cbmapget(attrs, "author", -1, NULL)) != NULL)
+            author = cbmemdup(tmp, -1);
+          if(!cdate && (tmp = cbmapget(attrs, "cdate", -1, NULL)) != NULL)
+            cdate = cbmemdup(tmp, -1);
+        }
+        cbmapclose(attrs);
+      } else {
+        if(!text && strisprintable(elem)){
+          rtext = strnormalize(elem, TRUE, TRUE);
+          text = cbxmlunescape(rtext);
+          free(rtext);
+        }
+      }
+    }
+    cblistclose(elems);
+    if(flags && author && cdate && text)
+      xsprintf(buf, "%s\n%s\n%s\n%s\n", flags, author, cdate, text);
+    free(text);
+    free(cdate);
+    free(author);
+    free(flags);
+  }
+  return cbdatumtomalloc(buf, NULL);
+}
+
+
 /* Get the jet lag by minute. */
 int jetlag(int min){
   static int lag = DEFJETLAG;
@@ -1429,6 +2030,75 @@ int matchcrypt(const char *key, const char *code){
 }
 
 
+/* Get the preamble for a LaTeX document. */
+const char *latexpreamble(const char *dclass, const char *title,
+                          const char *author, const char *date){
+  static char buf[CONSTBUFSIZ];
+  char *wp, *ptext;
+  wp = buf;
+  wp += sprintf(wp, "\\documentclass[a4papar]{%s}\n", dclass);
+  wp += sprintf(wp, "\\usepackage{graphicx}\n");
+  wp += sprintf(wp, "\\usepackage{verbatim}\n");
+  wp += sprintf(wp, "\\usepackage{color}\n");
+  wp += sprintf(wp, "\\usepackage{indentfirst}\n");
+  wp += sprintf(wp, "\\pagestyle{plain}\n");
+  wp += sprintf(wp, "\\pagenumbering{arabic}\n");
+  wp += sprintf(wp, "\\topmargin=-5mm\n");
+  wp += sprintf(wp, "\\oddsidemargin=-5mm\n");
+  wp += sprintf(wp, "\\evensidemargin=-5mm\n");
+  wp += sprintf(wp, "\\textwidth=170mm\n");
+  wp += sprintf(wp, "\\textheight=240mm\n");
+  wp += sprintf(wp, "\\parindent=1.0em\n");
+  wp += sprintf(wp, "\\parskip=0.7em\n");
+  wp += sprintf(wp, "\\renewcommand{\\baselinestretch}{1.0}\n");
+  wp += sprintf(wp, "\\let\\%s=\\verbatim\n", latexverbstr());
+  wp += sprintf(wp, "\\let\\end%s=\\endverbatim\n", latexverbstr());
+  wp += sprintf(wp, "\\definecolor{link}{rgb}{0.0,0.0,0.4}\n");
+  wp += sprintf(wp, "\\definecolor{ins}{rgb}{0.3,0.0,0.0}\n");
+  wp += sprintf(wp, "\\definecolor{del}{rgb}{0.3,0.3,0.3}\n");
+  ptext = strtolatex(title);
+  wp += sprintf(wp, "\\title{%s}\n", ptext);
+  free(ptext);
+  ptext = strtolatex(author);
+  wp += sprintf(wp, "\\author{%s}\n", ptext);
+  free(ptext);
+  ptext = strtolatex(date);
+  wp += sprintf(wp, "\\date{%s}\n", ptext);
+  free(ptext);
+  return buf;
+}
+
+
+/* Get a plain text in LaTeX format. */
+char *strtolatex(const char *str){
+  CBDATUM *buf;
+  char esc[NUMBUFSIZ];
+  int i;
+  assert(str);
+  buf = cbdatumopen("", 0);
+  for(i = 0; str[i] != '\0'; i++){
+    if(str[i] == '\\'){
+      cbdatumcat(buf, "{\\textbackslash}", -1);
+    } else if(str[i] == '<'){
+      cbdatumcat(buf, "{\\textless}", -1);
+    } else if(str[i] == '>'){
+      cbdatumcat(buf, "{\\textgreater}", -1);
+    } else if(str[i] == '|'){
+      cbdatumcat(buf, "{\\textbar}", -1);
+    } else if(strchr("^~`'-", str[i])){
+      sprintf(esc, "{\\char'%03o}", str[i]);
+      cbdatumcat(buf, esc, -1);
+    } else if(strchr("$&%#_{}", str[i])){
+      sprintf(esc, "\\%c", str[i]);
+      cbdatumcat(buf, esc, -1);
+    } else {
+      cbdatumcat(buf, str + i, 1);
+    }
+  }
+  return cbdatumtomalloc(buf, NULL);
+}
+
+
 
 /*************************************************************************************************
  * private objects
@@ -1471,11 +2141,11 @@ static char *uniqueid(const char *prefix){
 
 /* Check whether an ID string is valid.
    The return value is true if an ID string is valid, else, it is not. */
-static int validid(const char *id){
+static int validid(const char *id, const char *prefix){
   int i;
   assert(id);
-  if(!cbstrfwmatch(id, AIDPREFIX)) return FALSE;
-  for(i = strlen(AIDPREFIX); id[i] != '\0'; i++){
+  if(!cbstrfwmatch(id, prefix)) return FALSE;
+  for(i = strlen(prefix); id[i] != '\0'; i++){
     if(id[i] < '0' || id[i] > '9') return FALSE;
   }
   return TRUE;
@@ -1584,6 +2254,90 @@ static int strisprintable(const char *str){
 }
 
 
+/* Skip the label of a line.
+   `str' specifies a string with a label.
+   The return value is the pointer to the beginning of the label. */
+static const char *skiplabel(const char *str){
+  if(!(str = strchr(str, ':'))) return str;
+  str++;
+  while(*str != '\0' && (*str == ' ' || *str == '\t')){
+    str++;
+  }
+  return str;
+}
+
+
+/* Concatenate a Wiki-style text to an XML buffer.
+   `buf' specifies an XML buffer.
+   `text' specifies a text.
+   `trim' specifies whether to trim the text. */
+static void wikicat(CBDATUM *buf, const char *text, int trim){
+  const char *prefs[] = {
+    "http://", "https://", "ftp://", "gopher://", "ldap://", "file://", NULL
+  };
+  const char *pref, *pv, *rp;
+  char *rtext, *tmp, *mp;
+  int i, j;
+  rtext = cbmemdup(text, -1);
+  if(trim) cbstrtrim(rtext);
+  for(i = 0; rtext[i] != '\0'; i++){
+    if(cbstrfwmatch(rtext + i, "[[") && (pv = strstr(rtext + i, "]]")) != NULL){
+      tmp = cbmemdup(rtext + i + 2, pv - rtext - i - 2);
+      switch(tmp[0]){
+      case '^':
+        xsprintf(buf, "<emp>%@</emp>", tmp + 1);
+        break;
+      case '~':
+        xsprintf(buf, "<cite>%@</cite>", tmp + 1);
+        break;
+      case '+':
+        xsprintf(buf, "<ins>%@</ins>", tmp + 1);
+        break;
+      case '-':
+        xsprintf(buf, "<del>%@</del>", tmp + 1);
+        break;
+      default:
+        if((mp = strchr(tmp, '|')) != NULL){
+          *mp = '\0';
+          xsprintf(buf, "<link to=\"%@\">%@</link>", mp + 1, tmp);
+        } else {
+          xsprintf(buf, "%@", tmp);
+        }
+        break;
+      }
+      free(tmp);
+      i = pv - rtext + 1;
+    } else {
+      pref = NULL;
+      for(j = 0; prefs[j]; j++){
+        if(cbstrfwimatch(rtext + i, prefs[j])){
+          pref = prefs[j];
+        }
+      }
+      if(pref){
+        rp = rtext + i;
+        while(*rp != '\0' && ((*rp >= '0' && *rp <= '9') || (*rp >= 'A' && *rp <= 'Z') ||
+                              (*rp >= 'a' && *rp <= 'z') || strchr("-_.!~*'();/?:@&=+$,%#", *rp))){
+          rp++;
+        }
+        tmp = cbmemdup(rtext + i, rp - rtext - i);
+        xsprintf(buf, "<link to=\"%@\">%@</link>", tmp, tmp);
+        free(tmp);
+        i = rp - rtext - 1;
+      } else {
+        switch(rtext[i]){
+        case '&': xsprintf(buf, "&amp;"); break;
+        case '<': xsprintf(buf, "&lt;"); break;
+        case '>': xsprintf(buf, "&gt;"); break;
+        default: xsprintf(buf, "%c", rtext[i]); break;
+        }
+      }
+    }
+  }
+  free(rtext);
+}
+
+
 /* Extract HTML from the head of an article.
    `art' specifies a handle of a article object.
    The return value is a string of HTML. */
@@ -1652,6 +2406,8 @@ static char *bodytohtml(ARTICLE *art, const char *systemuri, int num){
         oname = name;
         if(!strcmp(name, "topic")){
           name = "h3";
+        } else if(!strcmp(name, "subtopic")){
+          name = "h4";
         } else if(!strcmp(name, "para")){
           name = "p";
         } else if(!strcmp(name, "asis")){
@@ -1677,8 +2433,9 @@ static char *bodytohtml(ARTICLE *art, const char *systemuri, int num){
         }
         if(elem[1] == '/'){
           xsprintf(buf, "</%@>", name);
-          if(!strcmp(name, "h3") || !strcmp(name, "p") || !strcmp(name, "pre") ||
-             !strcmp(name, "ul") || !strcmp(name, "li")) xsprintf(buf, "\n");
+          if(!strcmp(name, "h3") || !strcmp(name, "h4") || !strcmp(name, "p") ||
+             !strcmp(name, "pre") || !strcmp(name, "ul") || !strcmp(name, "li"))
+            xsprintf(buf, "\n");
           if(!strcmp(name, "ul")){
             inlist = FALSE;
           } else if(!strcmp(name, "li")){
@@ -1693,7 +2450,7 @@ static char *bodytohtml(ARTICLE *art, const char *systemuri, int num){
             xsprintf(buf, "<div class=\"%@\"><code>------</code></div>\n", oname);
           } else if(!strcmp(name, "a")){
             if(!(val = cbmapget(attrs, "to", -1, NULL))) val = "";
-            xsprintf(buf, "<a href=\"%@\" class=\"%@\">", val, oname);
+            xsprintf(buf, "<a href=\"%@\" rel=\"nofollow\" class=\"%@\">", val, oname);
           } else {
             xsprintf(buf, "<%@ class=\"%@\">", name, oname);
             if(!strcmp(name, "ul")){
@@ -1805,6 +2562,7 @@ const char *artsubclass(const char *flags){
   assert(flags);
   wp = name;
   if(strchr(flags, LONECHAR)) wp += sprintf(wp, " article-%s", LONESTR);
+  if(strchr(flags, WIKICHAR)) wp += sprintf(wp, " article-%s", WIKISTR);
   *wp = '\0';
   return name;
 }
@@ -1835,5 +2593,25 @@ static const char *hashclass(const char *prefix, const char *str){
 }
 
 
+/* Get the delimit string for vervatim sections of LaTeX.
+   The return value is the delimit string for vervatim sections of LaTeX. */
+static const char *latexverbstr(void){
+  static char buf[NUMBUFSIZ];
+  static int first = TRUE;
+  char *wp;
+  int i;
+  if(!first) return buf;
+  wp = buf;
+  srand(time(NULL));
+  wp += sprintf(wp, "verbatim");
+  for(i = 0; i < 8; i++){
+    *(wp++) = 'a' + rand() % ('z' - 'a');
+  }
+  *wp = '\0';
+  first = FALSE;
+  return buf;
+}
+
+
 
 /* END OF FILE */
