@@ -68,10 +68,13 @@
 #define MDATEKEY       "mdate"           /* key of the modification date */
 #define PATHBUFSIZ     2048              /* size of a path buffer */
 #define NUMBUFSIZ      32                /* size of a number buffer */
+#define CONSTBUFSIZ    4096              /* size of a buffer for const strings */
 #define DEFJETLAG      1440              /* default value of the jet lag */
 #define AIDPREFIX      "RA"              /* prefix of article ID */
 #define FIDPREFIX      "RF"              /* prefix of file ID */
 #define HASHNUM        7                 /* number of hash classes */
+#define WIKISTR        "wiki"            /* string for Wiki-style articles */
+#define WIKICHAR       'w'               /* character for Wiki-style articles */
 #define LONESTR        "lone"            /* string for lone articles */
 #define LONECHAR       'l'               /* character for lone articles */
 #define SINKSTR        "sink"            /* string for sinking responses */
@@ -238,15 +241,16 @@ int setattribute(RBBS *rbbs, const char *name, const char *value);
 
 /* Save a file into a database.
    `rbbs' specifies a database handle.
-   `password' specifies the password of a file.
+   `id' specifies the ID a file.  If it is `NULL', the ID is generated implicitly.
+   `password' specifies the password of the file.
    `subject' specifies the subject of the file.
    `author' specifies the author of the file.
    `cdate' specifies the creation date of the file.
    `ptr' specifies the pointer to the content of the file.
    `size' specifies the size of the content.
    The return value is true if successful, else it is false. */
-int savefile(RBBS *rbbs, const char *password, const char *subject, const char *author,
-             const char *cdate, const char *ptr, int size);
+int savefile(RBBS *rbbs, const char *id, const char *password, const char *subject,
+             const char *author, const char *cdate, const char *ptr, int size);
 
 
 /* Load a file in a database.
@@ -274,12 +278,24 @@ int deletefile(RBBS *rbbs, const char *id);
 CBLIST *listfiles(RBBS *rbbs);
 
 
-/* Make a object of article from an XML string.
+/* Get the number of files in a database.
+   `rbbs' specifies a database handle.
+   The return value is the number of files in the database. */
+int numfiles(RBBS *rbbs);
+
+
+/* Make an object of an article from an XML string.
    `xml' specifies a string of XML.
    The return value is a handle of an article object. */
 ARTICLE *makearticle(const char *xml);
 
 
+/* Make an object of an article from a wiki-style string.
+   `xml' specifies a wiki-style string.
+   The return value is a handle of an article object. */
+ARTICLE *makeartfromwiki(const char *wiki);
+
+
 /* Release resources of an article object.
    `art' specifies a handle of an article object. */
 void freearticle(ARTICLE *art);
@@ -416,6 +432,19 @@ char *arttohtml(ARTICLE *art, const char *systemuri, int bnum, int rnum);
 char *arttoatom(ARTICLE *art, const char *systemuri, int bnum, int rnum);
 
 
+/* Serialize an article into a string for LaTeX.
+   `art' specifies a handle of an article object.
+   `header' specifies whether to show headers of a chapter.
+   The return value is a string for LaTeX. */
+char *arttolatex(ARTICLE *art, int header);
+
+
+/* Serialize an article into a Wiki-style string.
+   `art' specifies a handle of an article object.
+   The return value is a Wiki-style string. */
+char *arttowiki(ARTICLE *art);
+
+
 /* Get the jet lag by minute.
    `min' specifies the jet lag to set.  If it is equal to or more than 1440, it is not set.
    The return value is the jet lag in minutes.  The default is the jet lag of the local time. */
@@ -476,6 +505,22 @@ const char *makecrypt(const char *key);
 int matchcrypt(const char *key, const char *code);
 
 
+/* Get the preamble for a LaTeX document.
+   `dclass' specifies the document class.
+   `title' specifies the title of a document.
+   `author' specifies the author of the document.
+   `date' specifies the author of the document.
+   The return value is the preamble of the document. */
+const char *latexpreamble(const char *dclass, const char *title,
+                          const char *author, const char *date);
+
+
+/* Get a plain text in LaTeX format.
+   `str' specifies a string.
+   The return value is a plain text in LaTeX format. */
+char *strtolatex(const char *str);
+
+
 
 #endif                                   /* duplication check */
 
